//*  Dropper Mutator v.2.0 (c) 2016 MrLoathsome aka Scott Armitage  *//
//*  Spawns Actors when players or monsters are killed.  *//
class Drop expands Mutator config(Dropper);

var() config int NumDrops;
var() config int SPNumDrops;
var() config int SPNumEx;
var() config bool bRandMode;
var() config bool bSPRandMode;
var() config bool bSuicideDrop;
var() config bool bMvsMspawn;
var() config bool bAnnounceSP;
var() config bool bRandMSizes;
var() config bool bDebugMode;
var() config float DropFreq;
var() config float SPDropFreq;
var() config float MaxMSize;
var() config float MinMSize;
var() config string DropList[255];
var() config float DLFreq[255];
var() config string DWK[255];
var() config int FRotate[255];
var() config float LSpan[255];
var() config float HMult[255];
var() config string SPDropList[255];
var() config string SPExclude[255];
var() config float SPDLFreq[255];
var() config string SPDWK[255];
var() config int SPFRotate[255];
var() config float SPLSpan[255];
var() config float SPHMult[255];

var int NumItems;
var int SPNumItems;

function PostBeginPlay()
{
	PreCacheLists();
	Super.PostBeginPlay();
}

function PreCacheLists()
{
	local int i;
	local bool bCWH;
	local class<Actor> aC;
	local Actor pcla;

	for (i=0; i<256; i++)
	{
		if (DropList[i] == "") break;
		aC = None;
		aC = class<Actor>(DynamicLoadObject(DropList[i], class'Class'));
		bCWH = aC.default.bCollideWorld;
		aC.default.bCollideWorld = False;
		pcla = spawn(aC,,,);
		aC.default.bCollideWorld = bCWH;
		if (pcla != None) pcla.destroy();
		if (DWK[i] != "")
		{
			aC = class<Actor>(DynamicLoadObject(DWK[i], class'Class'));
			bCWH = aC.default.bCollideWorld;
			aC.default.bCollideWorld = False;
			pcla = spawn(aC,,,);
			aC.default.bCollideWorld = bCWH;
			if (pcla != None) pcla.destroy();
		}
	}
	NumItems = i;
	for (i=0; i<256; i++)
	{
		if (SPDropList[i] == "") break;
		aC = None;
		aC = class<Actor>(DynamicLoadObject(SPDropList[i], class'Class'));
		bCWH = aC.default.bCollideWorld;
		aC.default.bCollideWorld = False;
		pcla = spawn(aC,,,);
		aC.default.bCollideWorld = bCWH;
		if (pcla != None) pcla.destroy();
		if (SPDWK[i] != "")
		{
			aC = class<Actor>(DynamicLoadObject(SPDWK[i], class'Class'));
			bCWH = aC.default.bCollideWorld;
			aC.default.bCollideWorld = False;
			pcla = spawn(aC,,,);
			aC.default.bCollideWorld = bCWH;
			if (pcla != None) pcla.destroy();
		}
	}
	SPNumItems = i;
	if (bDebugMode)
	{
		if (NumItems != 0) log("Dropper: PreCacheLists -"@NumItems@"valid DropList[] Classes found.");
		if (SPNumItems != 0) log("Dropper: PreCacheLists -"@SPNumItems@"valid SPDropList[] Classes found.");
	}
}

function bool HandlePickupQuery(Pawn Other, Inventory item, out byte bAllowPickup)
{
	item.LifeSpan = 0;
	if ( NextMutator != None )
		return NextMutator.HandlePickupQuery(Other, item, bAllowPickup);
	return false;
}

function bool PreventDeath(Pawn Killed, Pawn Killer, name damageType, vector HitLocation)
{
	if (((Killed == Killer) && (!bSuicideDrop)) || ((ScriptedPawn(Killed) != None) && (ScriptedPawn(Killer) != None) && (!bMvsMspawn)) || (Killed.Region.ZoneNumber == 0))
	{
		/* Skip */
	}
	else if ((DropFreq != 0.0) && (ScriptedPawn(Killed) == None) && (FRand() <= DropFreq))
		   { DropRandItems(Killed); }
	     else if ((SPDropFreq != 0.0) && (ScriptedPawn(Killed) != None) && (FRand() <= SPDropFreq))
			{ DropRandItemsSP(Killed); }

	if ( NextMutator != None )
		return NextMutator.PreventDeath(Killed, Killer, damageType, HitLocation);
	else return false;
}

function bool CheckCollision(Actor A, float DSM, vector newLoc)  // Credit to Feralidragon for help with this function
{
	local bool isColliding;
	local float CRadius, CHeight;

	CRadius = A.CollisionRadius;
	CHeight = A.CollisionHeight;
	A.SetCollisionSize(A.CollisionRadius * DSM, A.CollisionHeight * DSM);
	isColliding = !A.SetLocation(newLoc);
	A.SetCollisionSize(CRadius, CHeight);
	return isColliding;
}

function DropRandItems(Pawn Victim)
{
	local Vector V;
	local int rv, i, retrycnt, DHtmp;
	local float spawnfreq, F;
	local Actor A;
	local class<Actor> aClass;
	local pawn aV;
	local String DSetStr;

	retrycnt = 0;
	for (i=0; i<NumDrops; i++)
		{
			if (retrycnt == 0)
				{ if (bRandMode) { rv = (FRand()*NumItems); } else { rv = i; } }
			if (DLFreq[rv] != 0.0) { spawnfreq = DLFreq[rv]; } else { spawnfreq = 1.0; }
			if (FRand() <= spawnfreq)
			{
				if (retrycnt == 0) aClass = class<Actor>(DynamicLoadObject(DropList[rv], class'Class'));
				if ( aClass != None )
				{
					A = Spawn(aClass,,,Victim.Location+((aClass.default.CollisionHeight+aClass.default.CollisionRadius)*(retrycnt+1*i+2)*VRand()));
					if ((A != None) && (A.Region.ZoneNumber != 0))
					{
						retrycnt = 0;
						if (LSpan[rv] != 0.0) A.LifeSpan = LSpan[rv];
						A.Velocity += 1.5 * VSize(Victim.Velocity) * VRand();
						if (A.Velocity.Z < 0) A.Velocity.Z *= -1;
						A.Velocity.Z += FMax(FRand()*330, 60);
						A.bCollideWorld = True;
						A.SetCollision(True, True, True);
						A.SetPhysics(PHYS_Falling);
						if (FRotate[rv] != 0)
						{
							A.RotationRate.Yaw = (FRand() * 20000) + 500;
							if (FRotate[rv] == 2)
							{
								if (FRand() > 0.50) A.RotationRate.Pitch=250000*FRand() - 125000;
								A.RotationRate.Yaw=250000*FRand() - 125000;
								if (FRand() > 0.50) A.RotationRate.Roll=250000*FRand() - 125000;
							}
							if (Inventory(A) != None) { Inventory(A).bRotatingPickup = True; }
						} else if (ScriptedPawn(A) == None) A.RotationRate = Rotator(Vect(0,0,0));
						if (Decoration(A) != None)
						{
							Decoration(A).bPushable = True;
							Decoration(A).bWasCarried = False;
							A.DesiredRotation = RotRand();
							A.bRotateToDesired=False;
							A.bFixedRotationDir=True;
							DHtmp = Int(Decoration(A).GetPropertyText("Health"));
							Decoration(A).SetPropertyText("Health", String(DHtmp*HMult[rv]));
							if (DWK[rv] != "")
							{
								Decoration(A).Contents = class<Inventory>(DynamicLoadObject(DWK[rv], class'Class'));
							}
							A.Spawn(class'DRotSet',A);
						}
						if (inventory(A) != None)
						{
							A.bBounce = True;
							inventory(A).RespawnTime = 0.0;
							A.DesiredRotation = RotRand();
							A.bRotateToDesired=False;
							A.bFixedRotationDir=True;
							A.Inventory = None;
							if (pickup(A) != None)
							{
								pickup(A).bCanActivate = True;
								pickup(A).bAutoActivate = False;
							}
							if (weapon(A) != None) weapon(A).bWeaponStay = False;
							A.Spawn(class'DRotSet',A);
 						}
						if (ScriptedPawn(A) != None)
						{
							ScriptedPawn(A).bIsPlayer = False;
							ScriptedPawn(A).bIsBoss = False;
							if (DWK[rv] != "")
							{
								ScriptedPawn(A).DropWhenKilled = class<Inventory>(DynamicLoadObject(DWK[rv], class'Class'));
							}
							if (bRandMSizes)
							{
								F = FMax(MinMSize, (FRand()*MaxMSize));
								V = A.Location;
								if ((F > 1.0) && (CheckCollision(A, F, V))) { F = FMax(MinMSize, FRand()); }
								A.DrawScale = A.DrawScale * F;
								A.Mass = A.Mass * F;
								if (ScriptedPawn(A).Shadow != None)
									{ ScriptedPawn(A).Shadow.DrawScale = ScriptedPawn(A).Shadow.DrawScale * F; }
								ScriptedPawn(A).Health = ScriptedPawn(A).Health * F;
								A.SetCollisionSize((A.CollisionRadius * F), (A.CollisionHeight * F));
							}
							if (HMult[rv] != 0.0)
								{ ScriptedPawn(A).Health = ScriptedPawn(A).Health * HMult[rv]; }
							ScriptedPawn(A).gotostate('Hunting');
							if (!A.isA('Fly'))
							{
								ScriptedPawn(A).bHateWhenTriggered = True;
								ScriptedPawn(A).Intelligence = BRAINS_Human;
								ScriptedPawn(A).bCanStrafe = True;
								ScriptedPawn(A).CombatStyle = 1.0;
								foreach VisibleActors( class 'pawn' , aV,, A.Location)
								{
									if ((aV != None)  && (aV != A))
									{
								        	ScriptedPawn(A).Hated = aV;   
								          	ScriptedPawn(A).SetEnemy(aV);
										ScriptedPawn(A).enemy = aV;
										ScriptedPawn(A).gotostate('Attacking');
										if (bDebugMode)
											{ Log("Dropper:"$A.GetHumanName()$" attacks "$aV.GetHumanName()$"  -  "$ScriptedPawn(A).enemy); }
										break;
									}
								}
							}
							if (bAnnounceSP)
							{
								if (bDebugMode)
								{
									BroadcastMessage(A.GetHumanName()$" has spawned."$" Scale = "$F$" DrawScale = "$A.DrawScale$" Health = "$ScriptedPawn(A).Health$" Mass = "$A.Mass);
									log("Dropper:"$A.GetHumanName()$" has spawned."$" Scale = "$F$" DrawScale = "$A.DrawScale$" Health = "$ScriptedPawn(A).Health$" Mass = "$A.Mass);
								}
								else
									BroadcastMessage(A.GetHumanName()$" has spawned.");
							}
						}
						if (bDebugMode) log("Dropper: Spawned Actor = "$A);
					}
					else
					{
						retrycnt++;
						if (A != None) A.Destroy();
						if (bDebugMode) log("Dropper: Attempt #"$retrycnt$" failed for Class: "$DropList[rv]);
						if (retrycnt < 20) { i--; } else { retrycnt = 0; }
					}
				}
			}
			else { i--; }
		}
}

function DropRandItemsSP(Pawn Victim)
{
	local Vector V;
	local int rv, i, retrycnt, DHtmp;
	local Actor A;
	local float spawnfreq, F;
	local class<Actor> aClass;
	local class<Pawn> pClass;
	local pawn aV;
	local String DSetStr;

	for (i=0; i<SPNumEx; i++)
	{
		pClass = class<Pawn>(DynamicLoadObject(SPExclude[i], class'Class'));
		if (classischildof(Victim.class, pClass))
		{
			if (bDebugMode) log("Dropper:  (SP) Exclude for: "$pClass); 
			return;
		}
	}

	retrycnt = 0;
	for (i=0; i<SPNumDrops; i++)
		{
			if (retrycnt == 0)
				{ if (bSPRandMode) { rv = (FRand()*SPNumItems); } else { rv = i; } }
			if (SPDLFreq[rv] != 0.0) { spawnfreq = SPDLFreq[rv]; } else { spawnfreq = 1.0; }
			if (FRand() <= spawnfreq)
			{
				if (retrycnt == 0) aClass = class<Actor>(DynamicLoadObject(SPDropList[rv], class'Class'));
				if ( aClass != None )
				{
					A = Spawn(aClass,,,Victim.Location+((aClass.default.CollisionHeight+aClass.default.CollisionRadius)*(retrycnt+1*i+2)*VRand()));
					if ((A != None) && (A.Region.ZoneNumber != 0))
					{
						retrycnt = 0;
						if (SPLSpan[rv] != 0.0) A.LifeSpan = SPLSpan[rv];
						A.Velocity += 1.5 * VSize(Victim.Velocity) * VRand();
						if (A.Velocity.Z < 0) A.Velocity.Z *= -1;
						A.Velocity.Z += FMax(FRand()*330, 60);
						A.bCollideWorld = True;
						A.SetCollision(True, True, True);
						A.SetPhysics(PHYS_Falling);
						if (SPFRotate[rv] != 0)
						{
							A.RotationRate.Yaw = (FRand() * 20000) + 500;
							if (SPFRotate[rv] == 2)
							{
								if (FRand() > 0.50) A.RotationRate.Pitch=250000*FRand() - 125000;
								A.RotationRate.Yaw=250000*FRand() - 125000;
								if (FRand() > 0.50) A.RotationRate.Roll=250000*FRand() - 125000;
							}
							if (Inventory(A) != None) { Inventory(A).bRotatingPickup = True; }
						} else if (ScriptedPawn(A) == None) A.RotationRate = Rotator(Vect(0,0,0));
						if (Decoration(A) != None)
						{
							Decoration(A).bPushable = True;
							Decoration(A).bWasCarried = False;
							A.DesiredRotation = RotRand();
							A.bRotateToDesired=False;
							A.bFixedRotationDir=True;
							DHtmp = Int(Decoration(A).GetPropertyText("Health"));
							Decoration(A).SetPropertyText("Health", String(DHtmp*SPHMult[rv]));
							if (SPDWK[rv] != "")
							{
								Decoration(A).Contents = class<Inventory>(DynamicLoadObject(SPDWK[rv], class'Class'));
							}
							A.Spawn(class'DRotSet',A);
						}
						if (inventory(A) != None)
						{
							inventory(A).RespawnTime = 0.0;
							A.bBounce = True;
							A.DesiredRotation = RotRand();
							A.bRotateToDesired=False;
							A.bFixedRotationDir=True;
							A.Inventory = None;
							if (pickup(A) != None)
							{
								pickup(A).bCanActivate = True;
								pickup(A).bAutoActivate = False;
							}
							if (weapon(A) != None) weapon(A).bWeaponStay = False;
							A.Spawn(class'DRotSet',A);
						}
						if (ScriptedPawn(A) != None)
						{
							ScriptedPawn(A).bIsPlayer = False;
							ScriptedPawn(A).bIsBoss = False;
							if (SPDWK[rv] != "")
							{
								ScriptedPawn(A).DropWhenKilled = class<Inventory>(DynamicLoadObject(SPDWK[rv], class'Class'));
							}
							if (bRandMSizes)
							{
								F = FMax(MinMSize, (FRand()*MaxMSize));
								V = A.Location;
								if ((F > 1.0) && (CheckCollision(A, F, V))) { F = FMax(MinMSize, FRand()); }
								A.DrawScale = A.DrawScale * F;
								A.Mass = A.Mass * F;
								if (ScriptedPawn(A).Shadow != None)
									{ ScriptedPawn(A).Shadow.DrawScale = ScriptedPawn(A).Shadow.DrawScale * F; }
								ScriptedPawn(A).Health = ScriptedPawn(A).Health * F;
								A.SetCollisionSize((A.CollisionRadius * F), (A.CollisionHeight * F));
							}
							if (SPHMult[rv] != 0.0)
								{ ScriptedPawn(A).Health = ScriptedPawn(A).Health * SPHMult[rv]; }
							ScriptedPawn(A).gotostate('Hunting');
							if (!A.isA('Fly'))
							{
								ScriptedPawn(A).bHateWhenTriggered = True;
								ScriptedPawn(A).Intelligence = BRAINS_Human;
								ScriptedPawn(A).bCanStrafe = True;
								ScriptedPawn(A).CombatStyle = 1.0;
								foreach VisibleActors( class 'pawn' , aV,, A.Location)
								{
									if ((aV != None) && (aV != A))
									{
								        	ScriptedPawn(A).Hated = aV;
								          	ScriptedPawn(A).SetEnemy(aV);
										ScriptedPawn(A).enemy = aV;
										ScriptedPawn(A).gotostate('Attacking');
										if (bDebugMode)
											{ Log("Dropper: (SP) " $ A.GetHumanName() $ " attacks " $ aV.GetHumanName()$"  -  "$ScriptedPawn(A).enemy); }
										break;
									}
								}
							}
							if (bAnnounceSP)
							{
								if (bDebugMode)
								{
									BroadcastMessage(A.GetHumanName()$" has spawned."$" Scale = "$F$" DrawScale = "$A.DrawScale$" Health = "$ScriptedPawn(A).Health$" Mass = "$A.Mass);
									log("Dropper: (SP)"$A.GetHumanName()$" has spawned."$" Scale = "$F$" DrawScale = "$A.DrawScale$" Health = "$ScriptedPawn(A).Health$" Mass = "$A.Mass);
								}
								else
									BroadcastMessage(A.GetHumanName()$" has spawned.");
							}
						}
						if (bDebugMode) log("Dropper: Spawned Actor (SP) = "$A);
					}
					else
					{
						retrycnt++;
						if (A != None) A.Destroy();
						if (bDebugMode) log("Dropper: (SP) Attempt #"$retrycnt$" failed for Class: "$SPDropList[rv]);
						if (retrycnt < 20) { i--; } else { retrycnt = 0; }
					}
				}
			}
			else { i--; }
		}
}

defaultproperties
{
     NumItems=0
     NumDrops=0
     SPNumItems=0
     SPNumDrops=0
     SPNumEx=0
     bRandMode=True
     bSPRandMode=True
     bSuicideDrop=False
     bMvsMspawn=True
     bAnnounceSP=True
     bRandMSizes=True
     bDebugMode=False
     DropFreq=0.5
     SPDropFreq=0.5
     MaxMSize=1.5
     MinMSize=0.33
     DropList(0)=""
     DLFreq(0)=1.0
     DWK(0)=""
     FRotate(0)=0
     LSpan(0)=0.0
     HMult(0)=1.0
     SPDropList(0)=""
     SPExclude(0)=""
     SPDLFreq(0)=1.0
     SPDWK(0)=""
     SPFRotate(0)=0
     SPLSpan(0)=0.0
     SPHMult(0)=1.0
}
